# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""" Utilities """

import builtins
import sys


def safe_print(s):
    """Print with protection against UTF-8 encoding errors"""
    # A little bit of dance to be able to test this code
    print_f = getattr(builtins, "print")
    try:
        print_f(s)
    except UnicodeEncodeError:
        # failed to encode so let's do encoding while ignoring errors
        # to print at least something
        # explicit `or ascii` since somehow on buildbot it seemed to
        # return None
        s = s.encode(
            getattr(sys.stdout, 'encoding', 'ascii') \
            or 'ascii', errors='ignore') \
            if hasattr(s, 'encode') else s
        print_f(s.decode())



from os.path import join

from datalad.consts import (
    DATALAD_DOTDIR,
    DATALAD_GIT_DIR,
)


# Make use of those in datalad.metadata
OLDMETADATA_DIR = join(DATALAD_DOTDIR, 'meta')
OLDMETADATA_FILENAME = 'meta.json'

METADATA_DIR = join(DATALAD_DOTDIR, 'metadata')
WEB_META_DIR = join(DATALAD_GIT_DIR, 'metadata')

DATASET_METADATA_FILE = join(METADATA_DIR, 'dataset.json')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test implementation of class GitRepo

"""

from os.path import join as opj

from datalad.support.annexrepo import AnnexRepo

from datalad.tests.utils_pytest import (
    assert_repo_status,
    eq_,
    skip_if_adjusted_branch,
    with_tempfile,
)


@skip_if_adjusted_branch
@with_tempfile
@with_tempfile
def test_AnnexRepo_add_submodule(source_path=None, path=None):
    source = AnnexRepo(source_path, create=True)
    (source.pathobj / 'test-annex.dat').write_text("content")
    source.save('some')

    top_repo = AnnexRepo(path, create=True)

    top_repo.add_submodule('sub', name='sub', url=source_path)
    top_repo.commit('submodule added')
    eq_([s["gitmodule_name"] for s in top_repo.get_submodules_()],
        ['sub'])

    assert_repo_status(top_repo, annex=True)
    assert_repo_status(opj(path, 'sub'), annex=False)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from unittest.mock import patch
from io import StringIO
from tempfile import NamedTemporaryFile
from datalad.tests.utils_pytest import (
    assert_equal,
    skip_if_on_windows,
    SkipTest,
)

from datalad import __main__
from ..auto import AutomagicIO


# automagic IO is not supported on windows
@skip_if_on_windows
@patch.object(AutomagicIO, 'activate')
@patch('sys.stdout', new_callable=StringIO)
def test_main_run_a_script(stdout=None, mock_activate=None):
    try:
        from datalad.auto import AutomagicIO
        raise SkipTest(
            'Test not compatible with DataLad that provides AutomagicIO '
            'itself')
    except ModuleNotFoundError:
        pass
    f = NamedTemporaryFile()
    f.write('print("Running the script")\n'.encode())
    f.flush()
    __main__.main(['__main__.py', f.name])
    assert_equal(stdout.getvalue().rstrip(), "Running the script")
    # And we have "activated"
    mock_activate.assert_called_once_with()



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test image extractor"""

from unittest import SkipTest

from datalad.tests.utils_pytest import (
    assert_in,
    assert_repo_status,
    assert_result_count,
    assert_status,
    eq_,
    with_tempfile,
)

try:
    from PIL import Image
except ImportError as exc:
    raise SkipTest(
       "No PIL module available or it cannot be imported") from exc

from os.path import dirname
from os.path import join as opj
from shutil import copy

from datalad.api import Dataset



target = {
    "dcterms:SizeOrDuration": [4, 3],
    "color_mode": "3x8-bit pixels, true color",
    "type": "dctype:Image",
    "spatial_resolution(dpi)": [72, 72],
    "format": "JPEG (ISO 10918)"
}


@with_tempfile(mkdir=True)
def test_image(path=None):
    ds = Dataset(path).create()
    ds.config.add('datalad.metadata.nativetype', 'image', scope='branch')
    copy(
        opj(dirname(dirname(dirname(__file__))), 'tests', 'data', 'exif.jpg'),
        path)
    ds.save()
    assert_repo_status(ds.path)
    res = ds.aggregate_metadata()
    assert_status('ok', res)
    res = ds.metadata('exif.jpg')
    assert_result_count(res, 1)

    # from this extractor
    meta = res[0]['metadata']['image']
    for k, v in target.items():
        eq_(meta[k], v)

    assert_in('@context', meta)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Internal low-level interface to Git repositories

"""

import os.path as op

import logging
from os.path import (
    join as opj,
    exists,
    isabs,
    curdir,
)

import posixpath
import warnings

from datalad.utils import (
    Path,
    posix_relpath,
)

from datalad.support.gitrepo import (
    GitRepo,
)
from datalad.support.exceptions import (
    InvalidGitRepositoryError,
)
from datalad.core.local.repo import repo_from_path

lgr = logging.getLogger('datalad.gitrepo')


class DeprecatedGitRepoMethods(object):
    """This class is just a container for GitRepo methods that have been deprecated
    in -core. The can be moved here to rot in peace.
    """

    def add_submodule(self, path, name=None, url=None, branch=None):
        """Add a new submodule to the repository.

        This will alter the index as well as the .gitmodules file, but will not
        create a new commit.  If the submodule already exists, no matter if the
        configuration differs from the one provided, the existing submodule
        is considered as already added and no further action is performed.

        NOTE: This method does not work with submodules that use git-annex adjusted
              branches. Use Repo.save() instead.

        Parameters
        ----------
        path : str
          repository-relative path at which the submodule should be located, and
          which will be created as required during the repository initialization.
        name : str or None
          name/identifier for the submodule. If `None`, the `path` will be used
          as name.
        url : str or None
          git-clone compatible URL. If `None`, the repository is assumed to
          exist, and the url of the first remote is taken instead. This is
          useful if you want to make an existing repository a submodule of
          another one.
        branch : str or None
          name of branch to be checked out in the submodule. The given branch
          must exist in the remote repository, and will be checked out locally
          as a tracking branch. If `None`, remote HEAD will be checked out.
        """
        warnings.warn(
            'GitRepo.add_submodule() is deprecated and will be removed in '
            'a future release. Use the Dataset method save() instead.',
            DeprecationWarning
        )

        if name is None:
            name = Path(path).as_posix()
        cmd = ['submodule', 'add', '--name', name]
        if branch is not None:
            cmd += ['-b', branch]
        if url is None:
            # repo must already exist locally
            subm = repo_from_path(op.join(self.path, path))
            # check that it has a commit, and refuse
            # to operate on it otherwise, or we would get a bastard
            # submodule that cripples git operations
            if not subm.get_hexsha():
                raise InvalidGitRepositoryError(
                    'cannot add subdataset {} with no commits'.format(subm))
            # make an attempt to configure a submodule source URL based on the
            # discovered remote configuration
            remote, branch = subm.get_tracking_branch()
            url = subm.get_remote_url(remote) if remote else None

        if url is None:
            # had no luck with a remote URL
            if not isabs(path):
                # need to recode into a relative path "URL" in POSIX
                # style, even on windows
                url = posixpath.join(curdir, posix_relpath(path))
            else:
                url = path
        cmd += [url, Path(path).as_posix()]
        self.call_git(cmd)
        # record dataset ID if possible for comprehesive metadata on
        # dataset components within the dataset itself
        subm_id = GitRepo(op.join(self.path, path)).config.get(
            'datalad.dataset.id', None)
        if subm_id:
            self.call_git(
                ['config', '--file', '.gitmodules', '--replace-all',
                 'submodule.{}.datalad-id'.format(name), subm_id])
        # ensure supported setup
        from datalad.support.gitrepo import _fixup_submodule_dotgit_setup
        _fixup_submodule_dotgit_setup(self, path)
        # TODO: return value

    def deinit_submodule(self, path, **kwargs):
        """Deinit a submodule

        Parameters
        ----------
        path: str
            path to the submodule; relative to `self.path`
        kwargs:
            see `__init__`
        """
        warnings.warn(
            'GitRepo.deinit_submodule() is deprecated and will be removed in '
            'a future release. Use call_git() instead.',
            DeprecationWarning
        )
        from datalad.support.gitrepo import to_options
        self.call_git(['submodule', 'deinit'] + to_options(**kwargs),
                      files=[path])
        # TODO: return value

    def update_submodule(self, path, mode='checkout', init=False):
        """Update a registered submodule.

        This will make the submodule match what the superproject expects by
        cloning missing submodules and updating the working tree of the
        submodules. The "updating" can be done in several ways depending
        on the value of submodule.<name>.update configuration variable, or
        the `mode` argument.

        Parameters
        ----------
        path : str
          Identifies which submodule to operate on by it's repository-relative
          path.
        mode : {checkout, rebase, merge}
          Update procedure to perform. 'checkout': the commit recorded in the
          superproject will be checked out in the submodule on a detached HEAD;
          'rebase': the current branch of the submodule will be rebased onto
          the commit recorded in the superproject; 'merge': the commit recorded
          in the superproject will be merged into the current branch in the
          submodule.
        init : bool
          If True, initialize all submodules for which "git submodule init" has
          not been called so far before updating.
          Primarily provided for internal purposes and should not be used directly
          since would result in not so annex-friendly .git symlinks/references
          instead of full featured .git/ directories in the submodules
        """
        warnings.warn(
            'GitRepo.update_submodule() is deprecated and will be removed in '
            'a future release. Use the dataset method update() instead.',
            DeprecationWarning
        )
        if GitRepo.is_valid_repo(self.pathobj / path):
            subrepo = GitRepo(self.pathobj / path, create=False)
            subbranch = subrepo.get_active_branch() if subrepo else None
            try:
                subbranch_hexsha = subrepo.get_hexsha(subbranch) if subrepo else None
            except ValueError:
                if subrepo.commit_exists("HEAD"):
                    # Not what we thought it was. Reraise.
                    raise
                else:
                    raise ValueError(
                        "Cannot add submodule that has an unborn branch "
                        "checked out: {}"
                        .format(subrepo.path))

        else:
            subrepo = None
            subbranch = None
            subbranch_hexsha = None

        cmd = ['submodule', 'update', '--%s' % mode]
        if init:
            cmd.append('--init')
            subgitpath = opj(self.path, path, '.git')
            if not exists(subgitpath):
                # TODO:  wouldn't with --init we get all those symlink'ed .git/?
                # At least let's warn
                lgr.warning(
                    "Do not use update_submodule with init=True to avoid git creating "
                    "symlinked .git/ directories in submodules"
                )
            #  yoh: I thought I saw one recently but thought it was some kind of
            #  an artifact from running submodule update --init manually at
            #  some point, but looking at this code now I worry that it was not
        self.call_git(cmd, files=[path])

        if not init:
            return

        # track branch originally cloned, only if we had a valid repo at the start
        updated_subbranch = subrepo.get_active_branch() if subrepo else None
        if subbranch and not updated_subbranch:
            # got into 'detached' mode
            # trace if current state is a predecessor of the branch_hexsha
            lgr.debug(
                "Detected detached HEAD after updating submodule %s which was "
                "in %s branch before", self.path, subbranch)
            detached_hexsha = subrepo.get_hexsha()
            if subrepo.get_merge_base(
                    [subbranch_hexsha, detached_hexsha]) == detached_hexsha:
                # TODO: config option?
                # in all likely event it is of the same branch since
                # it is an ancestor -- so we could update that original branch
                # to point to the state desired by the submodule, and update
                # HEAD to point to that location
                lgr.info(
                    "Submodule HEAD got detached. Resetting branch %s to point "
                    "to %s. Original location was %s",
                    subbranch, detached_hexsha[:8], subbranch_hexsha[:8]
                )
                branch_ref = 'refs/heads/%s' % subbranch
                subrepo.update_ref(branch_ref, detached_hexsha)
                assert(subrepo.get_hexsha(subbranch) == detached_hexsha)
                subrepo.update_ref('HEAD', branch_ref, symbolic=True)
                assert(subrepo.get_active_branch() == subbranch)
            else:
                lgr.warning(
                    "%s has a detached HEAD since cloned branch %s has another common ancestor with %s",
                    subrepo.path, subbranch, detached_hexsha[:8]
                )
        # TODO: return value



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Metadata handling (parsing, storing, querying)"""

import datalad.interface.common_cfg as config_module
from datalad.support.constraints import (
    EnsureBool,
    EnsureChoice,
    EnsureInt,
)


# We have to add the metadata specific default
# to the general config definitions to support
# metadata commands that ask for this config.

config_module.definitions['datalad.search.default-mode'] = {
    'ui': (
        'question', {
            'title': 'Default search mode',
            'text': 'Label of the mode to be used by default'
        }
    ),
    'type': EnsureChoice('egrep', 'textblob', 'autofield'),
    'default': 'egrep',
}

config_module.definitions['datalad.search.index-default-documenttype'] = {
    'ui': (
        'question', {
            'title': 'Type of search index documents',
            'text': 'Labels of document types to include in a default search index'
        }
    ),
    'type': EnsureChoice('all', 'datasets', 'files'),
    'default': 'datasets',
}

config_module.definitions['datalad.search.indexercachesize'] = {
    'ui': (
        'question', {
            'title': 'Maximum cache size for search index (per process)',
            'text': 'Actual memory consumption can be twice as high as this value in MB (one process per CPU is used)'
        }
    ),
    'default': 256,
    'type': EnsureInt(),
}

config_module.definitions['datalad.metadata.maxfieldsize'] = {
    'ui': (
        'question', {
            'title': 'Maximum metadata field size',
            'text': 'Metadata fields exceeding this size (in bytes/chars) are excluded from metadata extractio'
        }
    ),
    'default': 100000,
    'type': EnsureInt(),
}

config_module.definitions['datalad.metadata.nativetype'] = {
    'ui': (
        'question', {
            'title': 'Native dataset metadata scheme',
            'text': 'Set this label to engage a particular metadata extraction parser'
        }
    ),
}

config_module.definitions['datalad.metadata.store-aggregate-content'] = {
    'ui': (
        'question', {
            'title': 'Aggregated content metadata storage',
            'text': 'If this flag is enabled, content metadata is aggregated into superdataset to allow for discovery of individual files. If disable unique content metadata values are still aggregated to enable dataset discovery'
        }
    ),
    'type': EnsureBool(),
    'default': True,
}

config_module.definitions['datalad.metadata.create-aggregate-annex-limit'] = {
    'ui': (
        'question', {
            'title': 'Limit configuration annexing aggregated metadata in new dataset',
            'text': 'Git-annex large files expression (see https://git-annex.branchable.com/tips/largefiles; given expression will be wrapped in parentheses)'
        }
    ),
    'default': 'anything',
}



# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test create publication target ssh web server action

"""

import re
from os.path import (
    exists,
    join as opj,
)
from datalad.utils import (
    _path_,
)
from datalad.tests.utils_pytest import (
    assert_false,
    assert_not_in,
    on_windows,
    ok_exists,
    ok_file_has_content,
    skip_if_adjusted_branch,
)


def assert_publish_with_ui(target_path, rootds=False, flat=True):
    paths = [_path_(".git/hooks/post-update")]     # hooks enabled in all datasets
    not_paths = []  # _path_(".git/datalad/metadata")]  # metadata only on publish
                    # ATM we run post-update hook also upon create since it might
                    # be a reconfiguration (TODO: I guess could be conditioned)

    # web-interface html pushed to dataset root
    web_paths = ['index.html', _path_(".git/datalad/web")]
    if rootds:
        paths += web_paths
    # and not to subdatasets
    elif not flat:
        not_paths += web_paths

    for path in paths:
        ok_exists(opj(target_path, path))

    for path in not_paths:
        assert_false(exists(opj(target_path, path)))

    hook_path = _path_(target_path, '.git/hooks/post-update')
    # No longer the case -- we are no longer using absolute path in the
    # script
    # ok_file_has_content(hook_path,
    #                     '.*\ndsdir="%s"\n.*' % target_path,
    #                     re_=True,
    #                     flags=re.DOTALL)
    # No absolute path (so dataset could be moved) in the hook
    with open(hook_path) as f:
        assert_not_in(target_path, f.read())
    # correct ls_json command in hook content (path wrapped in "quotes)
    ok_file_has_content(hook_path,
                        '.*datalad ls -a --json file \..*',
                        re_=True,
                        flags=re.DOTALL)


# we are simply running the tests from -core, which will behave differently
# when they find this extension to be installed. We are likely running to
# many tests for the webui scope, but that functionality is not cleanly
# separated in the tests -- but run too many than too few.

from datalad.distribution.tests.test_create_sibling import (
    test_invalid_call,
    test_target_ssh_recursive,
    test_target_ssh_since,
    test_replace_and_relative_sshpath,
    test_target_ssh_inherit,
    test_local_path_target_dir,
    test_local_relpath,
    test_non_master_branch,
    test_target_ssh_simple,
    test_failon_no_permissions,
    test_preserve_attrs,
    test_check_exists_interactive,
)


# DataLad core does not run crippled tests across the entire test suite
test_target_ssh_recursive = skip_if_adjusted_branch(test_target_ssh_recursive)
test_preserve_attrs = skip_if_adjusted_branch(test_preserve_attrs)


# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# -*- coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Some additional tests for search command (some are within test_base)"""

import logging
import os
from os import makedirs
from os.path import dirname
from os.path import join as opj
from pathlib import Path
from shutil import copy
from uuid import UUID
from pkg_resources import EntryPoint

from unittest.mock import (
    MagicMock,
    patch,
)

from datalad.api import (
    Dataset,
)
from datalad.support.exceptions import NoDatasetFound
from datalad.tests.utils_pytest import (
    SkipTest,
    assert_equal,
    assert_in,
    assert_is_generator,
    assert_raises,
    assert_re_in,
    assert_repo_status,
    assert_result_count,
    eq_,
    known_failure_githubci_win,
    ok_file_under_git,
    patch_config,
    with_tempfile,
    with_testsui,
)
from datalad.utils import (
    chpwd,
    swallow_logs,
    swallow_outputs,
)

from ..indexers.base import MetadataIndexer
from ..metadata import (
    NoMetadataStoreFound,
    gen4_query_aggregated_metadata,
    query_aggregated_metadata,
)
from ..search import (
    _listdict2dictlist,
    _meta2autofield_dict,
    Search
)

search = Search.__call__


@with_testsui(interactive=False)
@with_tempfile(mkdir=True)
def test_search_outside1_noninteractive_ui(tdir=None):
    # we should raise an informative exception
    with chpwd(tdir):
        with assert_raises(NoDatasetFound) as cme:
            list(search("irrelevant"))
        assert_in('run interactively', str(cme.value))


@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_search_outside1(tdir=None, newhome=None):
    with chpwd(tdir):
        # should fail since directory exists, but not a dataset
        # should not even waste our response ;)
        with patch_config({'datalad.locations.default-dataset': newhome}):
            gen = search("irrelevant", return_type='generator')
            assert_is_generator(gen)
            assert_raises(NoDatasetFound, next, gen)

        # and if we point to some non-existing dataset
        with assert_raises(ValueError):
            next(search("irrelevant", dataset=newhome))


@with_testsui(responses='yes')
@with_tempfile(mkdir=True)
@with_tempfile()
def test_search_outside1_install_default_ds(tdir=None, default_dspath=None):
    with chpwd(tdir):
        # let's mock out even actual install/search calls
        with \
            patch_config({'datalad.locations.default-dataset': default_dspath}), \
            patch('datalad.api.install',
                  return_value=Dataset(default_dspath)) as mock_install, \
            patch('datalad.distribution.dataset.Dataset.search',
                  new_callable=_mock_search):
            _check_mocked_install(default_dspath, mock_install)

            # now on subsequent run, we want to mock as if dataset already exists
            # at central location and then do search again
            from datalad.ui import ui
            ui.add_responses('yes')
            mock_install.reset_mock()
            with patch(
                    'datalad.distribution.dataset.Dataset.is_installed',
                    True):
                _check_mocked_install(default_dspath, mock_install)

            # and what if we say "no" to install?
            ui.add_responses('no')
            mock_install.reset_mock()
            with assert_raises(NoDatasetFound):
                list(search("."))

            # and if path exists and is a valid dataset and we say "no"
            Dataset(default_dspath).create()
            ui.add_responses('no')
            mock_install.reset_mock()
            with assert_raises(NoDatasetFound):
                list(search("."))


_mocked_search_results = [
    {
        'action': 'search',
        'status': 'ok',
        'path': 'ds1',  # this is wrong and must be an abspath
        'matched': {'f': 'v'},  # this has nothing to do with the actual output
    },
    {
        'action': 'search',
        'status': 'ok',
        'path': 'd2/ds2',  # this is wrong and must be an abspath
        'matched': {'f1': 'v1'}  # this has nothing to do with the actual output,
    },
]


class _mock_search(object):
    def __call__(*args, **kwargs):
        for report in _mocked_search_results:
            yield report


def _check_mocked_install(default_dspath, mock_install):
    gen = search(".", return_type='generator')
    assert_is_generator(gen)
    # we no longer do any custom path tune up from the one returned by search
    # so should match what search returns
    assert_equal(
        list(gen), [report
                    for report in _mocked_search_results])
    mock_install.assert_called_once_with(
        default_dspath,
        source='///',
        result_renderer='disabled')


@with_tempfile
def test_search_non_dataset(tdir=None):
    from datalad.support.gitrepo import GitRepo
    GitRepo(tdir, create=True)
    with assert_raises(NoDatasetFound) as cme:
        list(search('smth', dataset=tdir))
    # Should instruct user how that repo could become a datalad dataset
    assert_in("datalad create --force", str(cme.value))


@known_failure_githubci_win
@with_tempfile(mkdir=True)
def test_within_ds_file_search(path=None):
    try:
        import mutagen
    except ImportError:
        raise SkipTest
    ds = Dataset(path).create(force=True)
    # override default and search for datasets and files for this test
    for m in ('egrep', 'textblob', 'autofield'):
        ds.config.add(
            'datalad.search.index-{}-documenttype'.format(m), 'all',
            scope='branch')
    ds.config.add('datalad.metadata.nativetype', 'audio', scope='branch')
    makedirs(opj(path, 'stim'))
    for src, dst in (
            ('audio.mp3', opj('stim', 'stim1.mp3')),):
        copy(
            opj(dirname(dirname(__file__)), 'tests', 'data', src),
            opj(path, dst))
    ds.save()
    ok_file_under_git(path, opj('stim', 'stim1.mp3'), annexed=True)
    # If it is not under annex, below addition of metadata silently does
    # not do anything
    ds.repo.set_metadata(
        opj('stim', 'stim1.mp3'), init={'importance': 'very'})
    ds.aggregate_metadata()
    assert_repo_status(ds.path)
    # basic sanity check on the metadata structure of the dataset
    dsmeta = ds.metadata('.', reporton='datasets')[0]['metadata']
    for src in ('audio',):
        # something for each one
        assert_in(src, dsmeta)
        # each src declares its own context
        assert_in('@context', dsmeta[src])
        # we have a unique content metadata summary for each src
        assert_in(src, dsmeta['datalad_unique_content_properties'])

    # test default behavior
    with swallow_outputs() as cmo:
        ds.search(show_keys='name', mode='textblob')

        assert_in("""\
id
meta
parentds
path
type
""", cmo.out)

    target_out = """\
annex.importance
annex.key
audio.bitrate
audio.duration(s)
audio.format
audio.music-Genre
audio.music-album
audio.music-artist
audio.music-channels
audio.music-sample_rate
audio.name
audio.tracknumber
datalad_core.id
datalad_core.refcommit
id
parentds
path
type
"""

    # test default behavior while limiting set of keys reported
    with swallow_outputs() as cmo:
        ds.search([r'\.id', 'artist$'], show_keys='short')
        out_lines = [l for l in cmo.out.splitlines() if l]
        # test that only the ones matching were returned
        assert_equal(
            [l for l in out_lines if not l.startswith(' ')],
            ['audio.music-artist', 'datalad_core.id']
        )
        # more specific test which would also test formatting
        assert_equal(
            out_lines,
            ['audio.music-artist',
             ' in  1 datasets', " has 1 unique values: 'dlartist'",
             'datalad_core.id',
             ' in  1 datasets',
             # we have them sorted
             " has 1 unique values: '%s'" % ds.id
             ]
        )

    with assert_raises(ValueError) as cme:
        ds.search('*wrong')
    assert_re_in(
        r"regular expression '\(\?i\)\*wrong' \(original: '\*wrong'\) is incorrect: ",
        str(cme.value))

    # check generated autofield index keys
    with swallow_outputs() as cmo:
        ds.search(mode='autofield', show_keys='name')
        # it is impossible to assess what is different from that dump
        assert_in(target_out, cmo.out)

    assert_result_count(ds.search('blablob#'), 0)
    # now check that we can discover things from the aggregated metadata
    for mode, query, hitpath, matched in (
        ('egrep',
         ':mp3',
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # same as above, leading : is stripped, in indicates "ALL FIELDS"
        ('egrep',
         'mp3',
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # same as above, but with AND condition
        # get both matches
        ('egrep',
         ['mp3', 'type:file'],
         opj('stim', 'stim1.mp3'),
         {'type': 'file', 'audio.format': 'mp3'}),
        # case insensitive search
        ('egrep',
         'mp3',
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # field selection by expression
        ('egrep',
         r'audio\.+:mp3',
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # random keyword query
        ('textblob',
         'mp3',
         opj('stim', 'stim1.mp3'),
         {'meta': 'mp3'}),
        # report which field matched with auto-field
        ('autofield',
         'mp3',
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # XXX next one is not supported by current text field analyser
        # decomposes the mime type in [mime, audio, mp3]
        # ('autofield',
        # "'mime:audio/mp3'",
        # opj('stim', 'stim1.mp3'),
        # 'audio.format', 'mime:audio/mp3'),
        # but this one works
        ('autofield',
         "'mime audio mp3'",
         opj('stim', 'stim1.mp3'),
         {'audio.format': 'mp3'}),
        # TODO extend with more complex queries to test whoosh
        # query language configuration
    ):
        res = ds.search(query, mode=mode, full_record=True)
        assert_result_count(
            res, 1, type='file', path=opj(ds.path, hitpath),
            # each file must report the ID of the dataset it is from, critical for
            # discovering related content
            dsid=ds.id)
        # in egrep we currently do not search unique values
        # and the queries above aim at files
        assert_result_count(res, 1 if mode == 'egrep' else 2)
        if mode != 'egrep':
            assert_result_count(
                res, 1, type='dataset', path=ds.path, dsid=ds.id)
        # test the key and specific value of the match
        for matched_key, matched_val in matched.items():
            assert_in(matched_key, res[-1]['query_matched'])
            assert_equal(res[-1]['query_matched'][matched_key], matched_val)

    # test a suggestion msg being logged if no hits and key is a bit off
    with swallow_logs(new_level=logging.INFO) as cml:
        res = ds.search('audio.formats:mp3 audio.bitsrate:1', mode='egrep')
        assert not res
        assert_in('Did you mean any of', cml.out)
        assert_in('audio.format', cml.out)
        assert_in('audio.bitrate', cml.out)


def test_listdict2dictlist():
    f = _listdict2dictlist
    l1 = [1, 3, [1, 'a']]
    assert f(l1) is l1, "we return it as is if no emb dict"
    eq_(f([{1: 2}]), {1: 2})  # inside out no need for a list
    # inside out, join into the list, skip entry with a list, or space
    eq_(f([{1: [2, 3], 'a': 1}, {'a': 2, 'c': ''}]), {'a': [1, 2]})


def test_meta2autofield_dict():
    # Just a test that we would obtain the value stored for that extractor
    # instead of what unique values it already had (whatever that means)
    eq_(
        _meta2autofield_dict({
            'datalad_unique_content_properties':
                {'extr1': {"prop1": "v1"}},
            'extr1': {'prop1': 'value'}}),
        {'extr1.prop1': 'value'}
    )


def test_external_indexer():
    """ check that external indexer are called """
    class MockedIndexer(MetadataIndexer):
        def __init__(self, metadata_format_name: str):
            super().__init__(metadata_format_name)

        def create_index(self, metadata):
            yield from {
                "jubel": 1,
                "trubel": 2
            }.items()

    class MockedEntryPoint(EntryPoint):
        def __init__(self):
            pass

        def load(self, *args):
            return MockedIndexer

    def _mocked_iter_entry_points(group, metadata):
        yield MockedEntryPoint()

    with patch('pkg_resources.iter_entry_points',
               MagicMock(side_effect=_mocked_iter_entry_points)):
        index = _meta2autofield_dict({
            'datalad_unique_content_properties': {
                'extr1': {
                    "prop1": "v1"
                }
            },
            'extr1': {
                'prop1': 'value'
            }
        })

    eq_(
        index,
        {
            'extr1.jubel': '1',
            'extr1.trubel': '2'
        }
    )


def test_faulty_external_indexer():
    """ check that generic indexer is called on external indexer faults """
    class MockedEntryPoint(EntryPoint):
        def __init__(self):
            self.name = 'MockedEntryPoint'
            self.dist = 'Mock Distribution 1.1'

        def load(self, *args):
            raise Exception('Mocked indexer error')

    def _mocked_iter_entry_points(group, metadata):
        yield MockedEntryPoint()

    with patch('pkg_resources.iter_entry_points',
               MagicMock(side_effect=_mocked_iter_entry_points)):

        index = _meta2autofield_dict({
            'datalad_unique_content_properties': {
                'extr1': {
                    "prop1": "v1"
                }
            },
            'extr1': {
                'prop1': 'value'
            }
        })

    eq_(
        index,
        {
            'extr1.prop1': 'value'
        }
    )


def test_multiple_entry_points():
    """ check that generic indexer is called if multiple indexers exist for the same name """
    class MockedEntryPoint(EntryPoint):
        def __init__(self):
            self.name = 'MockedEntryPoint'
            self.dist = 'Mock Distribution 1.1'

        def load(self, *args):
            return 'Loaded MockedEntryPoint'

    def _mocked_iter_entry_points(group, metadata):
        yield MockedEntryPoint()
        yield MockedEntryPoint()

    with patch('pkg_resources.iter_entry_points',
               MagicMock(side_effect=_mocked_iter_entry_points)):

        index = _meta2autofield_dict({
            'datalad_unique_content_properties': {
                'extr1': {
                    "prop1": "v1"
                }
            },
            'extr1': {
                'prop1': 'value'
            }
        })

    eq_(
        index,
        {
            'extr1.prop1': 'value'
        }
    )


def test_gen4_query_aggregated_metadata():
    class DatasetMock:
        def __init__(self, path: str):
            self.path = path
            self.pathobj = Path(path)

    extractor_name = 'mocked_extractor'
    extracted_metadata = {'info': 'this is mocked metadata'}

    def mocked_dump(**kwargs):
        yield from [
            {
                'action': 'dump',
                'status': 'ok',
                'path': '/ds1',
                'metadata': {
                    'type': 'file',
                    'dataset_id': str(UUID(int=1234)),
                    'dataset_version': '0000001111111',
                    'extractor_name': extractor_name,
                    'path': kwargs['path'],
                    'extracted_metadata': extracted_metadata
                }
            },
            {
                'action': 'dump',
                'status': 'error',
                'path': '/ds1',
                'metadata': {
                    'type': 'file',
                    'dataset_id': str(UUID(int=12345)),
                    'dataset_version': '0000002222222',
                    'extractor_name': extractor_name,
                    'path': kwargs['path'],
                    'extracted_metadata': extracted_metadata
                }
            },
            {
                'action': 'dump',
                'status': 'ok',
                'path': '/ds1',
                'metadata': {
                    'type': 'something',
                    'dataset_id': str(UUID(int=123456)),
                    'dataset_version': '0000003333333',
                    'extractor_name': extractor_name,
                    'path': kwargs['path'],
                    'extracted_metadata': extracted_metadata
                }
            }
        ]

    def mocked_no_metadata(**kwargs):
        raise NoMetadataStoreFound()

    mocked_annotated_paths = [
        {
            'path': 'ds/p1',
            'types': ['file', 'dataset']
        }
    ]

    with patch("datalad_deprecated.metadata.metadata.Dump") as dump_mock:
        dump_mock.return_value = mocked_dump

        result = tuple(
            gen4_query_aggregated_metadata(
                reporton='all',
                ds=DatasetMock('ds'),
                aps=mocked_annotated_paths
            )
        )
        assert_equal(len(result), 1)
        assert_equal(
            result[0]['metadata'],
            {extractor_name: extracted_metadata}
        )

        # ensure gen4 code path selection
        import datalad_deprecated.metadata.metadata
        datalad_deprecated.metadata.metadata.next_generation_metadata_available = True

        result = [
            metadata_result
            for metadata_result in query_aggregated_metadata(
                reporton='all',
                ds=DatasetMock('ds'),
                aps=mocked_annotated_paths,
                metadata_source='gen4'
            )
            if metadata_result['status'] == 'ok'
        ]
        assert len(result) == 1
        assert result[0]['metadata'] == {extractor_name: extracted_metadata}

        # check no metadata found-handling
        dump_mock.reset_mock()
        dump_mock.return_value = mocked_no_metadata

        result = [
            metadata_result
            for metadata_result in query_aggregated_metadata(
                reporton='all',
                ds=DatasetMock('ds'),
                aps=mocked_annotated_paths,
                metadata_source='all',
            )
            if metadata_result['status'] == 'impossible'
        ]
        assert_equal(len(result), 2)


@with_tempfile(mkdir=True)
def test_metadata_source_handling(temp_dir=None):
    import datalad_deprecated.metadata.metadata as metadata_module

    temp_ds = Dataset(temp_dir).create()

    gen4_result, legacy_result = (
        {
            "status": "ok",
            "path": "/ds",
            "action": "search",
            "matches": {("x", "x"): True},
            "metadata": {"k1": "v1", "k2": "v2"},
            "metadata_source": source
        }
        for source in ("gen4", "legacy")
    )

    def reset_mock(mock, result):
        mock.reset_mock()
        mock.return_value = [result]

    metadata_module.next_generation_metadata_available = True
    with patch("datalad_deprecated.metadata.metadata.gen4_query_aggregated_metadata") as gen4_mock, \
         patch("datalad_deprecated.metadata.metadata.legacy_query_aggregated_metadata") as legacy_mock:

        reset_mock(gen4_mock, gen4_result)
        reset_mock(legacy_mock, legacy_result)
        r = tuple(
            result["metadata_source"]
            for result in search(dataset=temp_ds, query="v1", metadata_source='all')
        )
        assert_equal(r, ("legacy", "gen4"))

        reset_mock(gen4_mock, gen4_result)
        reset_mock(legacy_mock, legacy_result)
        r = tuple(
            result["metadata_source"]
            for result in search(dataset=temp_ds, query="v1", metadata_source="gen4")
        )
        assert_equal(r, ("gen4",))

        reset_mock(gen4_mock, gen4_result)
        reset_mock(legacy_mock, legacy_result)
        r = tuple(
            result["metadata_source"]
            for result in search(dataset=temp_ds, query="v1", metadata_source="legacy")
        )
        assert_equal(r, ("legacy",))



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test implementation of class GitRepo

"""

import logging

import os.path as op

import pytest

from datalad.tests.utils_pytest import (
    assert_false,
    assert_in,
    assert_raises,
    assert_repo_status,
    DEFAULT_BRANCH,
    eq_,
    ok_,
    SkipTest,
    swallow_logs,
    with_tempfile,
)

# we nevertheless import from -core
# which must provide all deprecated methods whenever
# this package is installed too
from datalad.support.gitrepo import GitRepo
from datalad.support.exceptions import (
    CommandError,
)


@with_tempfile
@with_tempfile
@with_tempfile
def test_submodule_deinit(src=None, subsrc=None, path=None):
    src = GitRepo(src)
    subsrc = GitRepo(subsrc)
    for repo in (src, subsrc):
        for filename in ('some1.txt', 'some2.dat'):
            with open(op.join(repo.path, filename), 'w') as f:
                f.write(filename)
            repo.add(filename)
        repo.commit('Some files')
    src.add_submodule('subm 1', name='subm 1', url=subsrc.path)
    src.add_submodule('2', name='2', url=subsrc.path)
    src.commit('submodule added')

    top_repo = GitRepo.clone(src.path, path)
    eq_({'subm 1', '2'},
        {s["gitmodule_name"] for s in top_repo.get_submodules_()})
    # note: here init=True is ok, since we are using it just for testing
    with swallow_logs(new_level=logging.WARN) as cml:
        top_repo.update_submodule('subm 1', init=True)
        assert_in('Do not use update_submodule with init=True', cml.out)
    top_repo.update_submodule('2', init=True)

    # ok_(all([s.module_exists() for s in top_repo.get_submodules()]))
    # TODO: old assertion above if non-bare? (can't use "direct mode" in test_gitrepo)
    # Alternatively: New testrepo (plain git submodules) and have a dedicated
    # test for annexes in addition
    ok_(all(GitRepo.is_valid_repo(s["path"])
            for s in top_repo.get_submodules_()))

    # modify submodule:
    with open(op.join(top_repo.path, 'subm 1', 'file_ut.dat'), "w") as f:
        f.write("some content")

    assert_raises(CommandError, top_repo.deinit_submodule, 'sub1')

    # using force should work:
    top_repo.deinit_submodule('subm 1', force=True)

    ok_(not GitRepo.is_valid_repo(str(top_repo.pathobj / 'subm 1')))


@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_GitRepo_add_submodule(source_path=None, path=None):
    source = GitRepo(source_path, create=True)
    with open(op.join(source_path, 'some.txt'), 'w') as f:
        f.write("New text file.")
    source.add('some.txt')
    source.commit('somefile')

    top_repo = GitRepo(path, create=True)

    top_repo.add_submodule('sub', name='sub', url=source_path)
    top_repo.commit('submodule added')
    eq_([s["gitmodule_name"] for s in top_repo.get_submodules_()],
        ['sub'])
    assert_repo_status(path)
    assert_repo_status(op.join(path, 'sub'))


def test_GitRepo_update_submodule():
    raise SkipTest("TODO")


@pytest.mark.parametrize("is_ancestor", [False, True])
@with_tempfile(mkdir=True)
def test_update_submodule_init_adjust_branch(path=None, *, is_ancestor):
    src = GitRepo(op.join(path, "src"), create=True)
    src_sub = GitRepo(op.join(src.path, "sub"), create=True)
    src_sub.commit(msg="c0", options=["--allow-empty"])
    src_sub.commit(msg="c1", options=["--allow-empty"])
    src.add_submodule('sub', name='sub')
    src.commit(msg="Add submodule")

    # Move subdataset past the registered commit...
    hexsha_registered = src_sub.get_hexsha()
    if is_ancestor:
        # ... where the registered commit is an ancestor of the new one.
        src_sub.commit(msg="c2", options=["--allow-empty"])
    else:
        # ... where the registered commit is NOT an ancestor of the new one.
        src_sub.call_git(["reset", "--hard", DEFAULT_BRANCH + "~1"])  # c0
    hexsha_sub = src_sub.get_hexsha()

    clone = GitRepo.clone(url=src.path,
                          path=op.join(path, "clone"),
                          create=True)
    clone_sub = GitRepo.clone(url=src_sub.path,
                              path=op.join(clone.path, "sub"),
                              create=True)
    ok_(clone.dirty)
    eq_(clone_sub.get_active_branch(), DEFAULT_BRANCH)
    eq_(hexsha_sub, clone_sub.get_hexsha())

    clone.update_submodule("sub", init=True)

    assert_false(clone.dirty)
    eq_(hexsha_registered, clone_sub.get_hexsha())
    if is_ancestor:
        eq_(clone_sub.get_active_branch(), DEFAULT_BRANCH)
    else:
        assert_false(clone_sub.get_active_branch())


@with_tempfile
def test_update_submodules_sub_on_unborn_branch(path=None):
    repo = GitRepo(path, create=True)
    repo.commit(msg="c0", options=["--allow-empty"])
    subrepo = GitRepo(op.join(path, "sub"), create=True)
    subrepo.commit(msg="s c0", options=["--allow-empty"])
    repo.add_submodule(path="sub")
    subrepo.checkout("other", options=["--orphan"])
    with assert_raises(ValueError) as cme:
        repo.update_submodule(path="sub")
    assert_in("unborn branch", str(cme.value))


@with_tempfile
def test_GitRepo_get_submodules(path=None):
    repo = GitRepo(path, create=True)

    s_abc = GitRepo(op.join(path, "s_abc"), create=True)
    s_abc.commit(msg="c s_abc", options=["--allow-empty"])
    repo.add_submodule(path="s_abc")

    s_xyz = GitRepo(op.join(path, "s_xyz"), create=True)
    s_xyz.commit(msg="c s_xyz", options=["--allow-empty"])
    repo.add_submodule(path="s_xyz")

    eq_([s["gitmodule_name"]
         for s in repo.get_submodules(sorted_=True)],
        ["s_abc", "s_xyz"])


@with_tempfile
def test_get_submodules_parent_on_unborn_branch(path=None):
    repo = GitRepo(path, create=True)
    subrepo = GitRepo(op.join(path, "sub"), create=True)
    subrepo.commit(msg="s", options=["--allow-empty"])
    repo.add_submodule(path="sub")
    eq_([s["gitmodule_name"] for s in repo.get_submodules_()],
        ["sub"])



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# -*- coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test metadata extraction"""

from os.path import dirname
from os.path import join as opj
from shutil import copy
from unittest import SkipTest

try:
    import libxmp
except Exception as e:
    raise SkipTest(f"Module 'libxmp' failed to load: {e}")

from datalad.coreapi import Dataset
from datalad.tests.utils_pytest import (
    assert_in,
    assert_raises,
    assert_repo_status,
    assert_result_count,
    skip_if_adjusted_branch,
    #skip_if_on_windows,
    with_tempfile,
)
from datalad.utils import chpwd
from datalad_deprecated.metadata.extract_metadata import ExtractMetadata

extract_metadata = ExtractMetadata.__call__
testpath = opj(dirname(dirname(dirname(__file__))), 'metadata', 'tests', 'data', 'xmp.pdf')


@with_tempfile(mkdir=True)
def test_error(path=None):
    # go into virgin dir to avoid detection of any dataset
    with chpwd(path):
        assert_raises(ValueError, extract_metadata, types=['bogus__'], files=[testpath])


@skip_if_adjusted_branch
@with_tempfile(mkdir=True)
def test_ds_extraction(path=None):
    ds = Dataset(path).create()
    copy(testpath, path)
    ds.save()
    assert_repo_status(ds.path)

    res = extract_metadata(
        types=['xmp'],
        dataset=ds,
        # artificially disable extraction from any file in the dataset
        files=[])
    assert_result_count(
        res, 1,
        type='dataset', status='ok', action='metadata', path=path, refds=ds.path)
    assert_in('xmp', res[0]['metadata'])

    # now the more useful case: getting everything for xmp from a dataset
    res = extract_metadata(
        types=['xmp'],
        dataset=ds)
    assert_result_count(res, 2)
    assert_result_count(
        res, 1,
        type='dataset', status='ok', action='metadata', path=path, refds=ds.path)
    assert_result_count(
        res, 1,
        type='file', status='ok', action='metadata', path=opj(path, 'xmp.pdf'),
        parentds=ds.path)
    for r in res:
        assert_in('xmp', r['metadata'])


@skip_if_adjusted_branch
@with_tempfile(mkdir=True)
def test_file_extraction(path=None):
    # go into virgin dir to avoid detection of any dataset
    with chpwd(path):
        res = extract_metadata(
            types=['xmp'],
            files=[testpath])
        assert_result_count(res, 1, type='file', status='ok', action='metadata', path=testpath)
        assert_in('xmp', res[0]['metadata'])



from datalad.support.json_py import load_stream
from datalad.utils import ensure_unicode


unicode_srctypes = str, bytes


def all_same(items):
    """Quick check if all items are the same.

    Identical to a check like len(set(items)) == 1 but
    should be more efficient while working on generators, since would
    return False as soon as any difference detected thus possibly avoiding
    unnecessary evaluations
    """
    first = True
    first_item = None
    for item in items:
        if first:
            first = False
            first_item = item
        else:
            if item != first_item:
                return False
    # So we return False if was empty
    return not first


def as_unicode(val, cast_types=object):
    """Given an arbitrary value, would try to obtain unicode value of it

    For unicode it would return original value, for python2 str or python3
    bytes it would use ensure_unicode, for None - an empty (unicode) string,
    and for any other type (see `cast_types`) - would apply the unicode
    constructor.  If value is not an instance of `cast_types`, TypeError
    is thrown

    Parameters
    ----------
    cast_types: type
      Which types to cast to unicode by providing to constructor
    """
    if val is None:
        return u''
    elif isinstance(val, str):
        return val
    elif isinstance(val, unicode_srctypes):
        return ensure_unicode(val)
    elif isinstance(val, cast_types):
        return str(val)
    else:
        raise TypeError(
            "Value %r is not of any of known or provided %s types"
            % (val, cast_types))


def load_xzstream(fname):
    yield from load_stream(fname, compressed=True)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# -*- coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test utilities

"""

import builtins
from unittest.mock import patch

from datalad.tests.utils_pytest import assert_equal

from datalad_deprecated.utils import safe_print


def test_safe_print():
    """Just to test that we are getting two attempts to print"""

    called = [0]

    def _print(s):
        assert_equal(s, "bua")
        called[0] += 1
        if called[0] == 1:
            raise UnicodeEncodeError('crap', u"", 0, 1, 'whatever')

    with patch.object(builtins, 'print', _print):
        safe_print("bua")
    assert_equal(called[0], 2)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Legacy code to upload a web UI to any SSH host sibling
"""

__docformat__ = 'restructuredtext'

from glob import glob
import logging
from os.path import (
    dirname,
    join as opj,
)

from datalad.consts import DATALAD_GIT_DIR
from datalad.distribution.create_sibling import mkdir_cmd
from datalad.support.sshconnector import sh_quote
from datalad.utils import (
    make_tempfile,
)
import datalad_deprecated

lgr = logging.getLogger('datalad.deprecated.sibling_webui')


WEB_HTML_DIR = opj(DATALAD_GIT_DIR, 'web')


def upload_web_interface(path, ssh, shared, ui):
    # path to web interface resources on local
    webui_local = opj(
        dirname(datalad_deprecated.__file__), 'resources', 'website')
    # local html to dataset
    html_local = opj(webui_local, "index.html")

    # name and location of web-interface html on target
    html_targetname = {True: ui, False: "index.html"}[isinstance(ui, str)]
    html_target = opj(path, html_targetname)

    # upload ui html to target
    ssh.put(html_local, html_target)

    # upload assets to the dataset
    webresources_local = opj(webui_local, 'assets')
    webresources_remote = opj(path, WEB_HTML_DIR)
    ssh('{} {}'.format(mkdir_cmd, sh_quote(webresources_remote)))
    ssh.put(webresources_local, webresources_remote, recursive=True)

    # minimize and upload js assets
    for js_file in glob(opj(webresources_local, 'js', '*.js')):
        with open(js_file) as asset:
            try:
                from jsmin import jsmin
                # jsmin = lambda x: x   # no minimization
                minified = jsmin(asset.read())                      # minify asset
            except ImportError:
                lgr.warning(
                    "Will not minify web interface javascript, no jsmin available")
                minified = asset.read()                             # no minify available
            with make_tempfile(content=minified) as tempf:          # write minified to tempfile
                js_name = js_file.split('/')[-1]
                ssh.put(tempf, opj(webresources_remote, 'assets', 'js', js_name))  # and upload js

    # explicitly make web+metadata dir of dataset world-readable, if shared set to 'all'
    mode = None
    if shared in (True, 'true', 'all', 'world', 'everybody'):
        mode = 'a+rX'
    elif shared == 'group':
        mode = 'g+rX'
    elif str(shared).startswith('0'):
        mode = shared

    if mode:
        ssh('chmod -R {} {} {}'.format(
            mode,
            sh_quote(dirname(webresources_remote)),
            sh_quote(opj(path, 'index.html'))))



import logging
import pytest
import re
import tempfile

from functools import wraps
from os.path import join as opj

from datalad import cfg as dl_cfg
from datalad.api import create
from datalad.cmd import (
    GitWitlessRunner,
    KillOutput,
)
from datalad.support.gitrepo import GitRepo
from datalad.support.annexrepo import AnnexRepo
from datalad.utils import (
    optional_args,
    is_explicit_path,
    get_tempfile_kwargs,
)
from datalad.utils import (
    rmtemp,
)
from datalad.tests.utils_pytest import (
    assert_false,
    attr,
    _get_resolved_flavors,
    create_tree,
)

from datalad.tests import _TEMP_PATHS_GENERATED

lgr = logging.getLogger("datalad.tests.utils_pytest")
_TESTREPOS = None
local_testrepo_flavors = ['local'] # 'local-url'
_TEMP_PATHS_CLONES = set()

if dl_cfg.obtain('datalad.tests.setup.testrepos'):
    lgr.debug("Pre-populating testrepos")
    from datalad_deprecated.tests.utils import with_testrepos

    with_testrepos()(lambda repo: 1)()


@optional_args
def with_testrepos(t, regex='.*', flavors='auto', skip=False, count=None):
    """Decorator to provide a local/remote test repository

    All tests under datalad/tests/testrepos are stored in two-level hierarchy,
    where top-level name describes nature/identifier of the test repository,
    and there could be multiple instances (e.g. generated differently) of the
    same "content"

    Parameters
    ----------
    regex : string, optional
      Regex to select which test repos to use
    flavors : {'auto', 'local', 'local-url', 'clone', 'network', 'network-clone'} or list of thereof, optional
      What URIs to provide.  E.g. 'local' would just provide path to the
      repository, while 'network' would provide url of the remote location
      available on Internet containing the test repository.  'clone' would
      clone repository first to a temporary location. 'network-clone' would
      first clone from the network location. 'auto' would include the list of
      appropriate ones (e.g., no 'network*' flavors if network tests are
      "forbidden").
    count: int, optional
      If specified, only up to that number of repositories to test with

    Examples
    --------

    >>> from datalad_deprecated.tests.utils import with_testrepos
    >>> @with_testrepos('basic_annex')
    ... def test_write(repo):
    ...    assert(os.path.exists(os.path.join(repo, '.git', 'annex')))

    """
    @wraps(t)
    @attr('with_testrepos')
    def  _wrap_with_testrepos(*arg, **kw):
        # addurls with our generated file:// URLs doesn't work on appveyor
        # https://ci.appveyor.com/project/mih/datalad/builds/29841505/job/330rwn2a3cvtrakj
        #if 'APPVEYOR' in os.environ:
        #    pytest.skip("Testrepo setup is broken on AppVeyor")
        # TODO: would need to either avoid this "decorator" approach for
        # parametric tests or again aggregate failures like sweepargs does
        flavors_ = _get_resolved_flavors(flavors)

        testrepos_uris = _get_testrepos_uris(regex, flavors_)
        # we should always have at least one repo to test on, unless explicitly only
        # network was requested by we are running without networked tests
        if not (dl_cfg.get('datalad.tests.nonetwork') and flavors == ['network']):
            assert(testrepos_uris)
        else:
            if not testrepos_uris:
                pytest.skip("No non-networked repos to test on")

        fake_dates = dl_cfg.get("datalad.fake-dates")
        ntested = 0
        for uri in testrepos_uris:
            if count and ntested >= count:
                break
            ntested += 1
            if __debug__:
                lgr.debug('Running %s on %s', t.__name__, uri)
            try:
                t(*(arg + (uri,)), **kw)
            finally:
                # The is_explicit_path check is needed because it may be a URL,
                # but check_dates needs a local path or GitRepo object.
                if fake_dates and is_explicit_path(uri):
                    from datalad.support.repodates import check_dates
                    assert_false(
                        check_dates(uri, annex="tree")["objects"])
                if uri in _TEMP_PATHS_CLONES:
                    _TEMP_PATHS_CLONES.discard(uri)
                    rmtemp(uri)
                pass  # might need to provide additional handling so, handle
    return  _wrap_with_testrepos
with_testrepos.__test__ = False



def _get_testrepos_uris(regex, flavors):
    global _TESTREPOS
    # we should instantiate those whenever test repos actually asked for
    # TODO: just absorb all this lazy construction within some class
    if not _TESTREPOS:
        from datalad.tests.utils_testrepos import (
            BasicAnnexTestRepo,
            BasicGitTestRepo,
            InnerSubmodule,
            NestedDataset,
            SubmoduleDataset,
        )

        _basic_annex_test_repo = BasicAnnexTestRepo()
        _basic_git_test_repo = BasicGitTestRepo()
        _submodule_annex_test_repo = SubmoduleDataset()
        _nested_submodule_annex_test_repo = NestedDataset()
        _inner_submodule_annex_test_repo = InnerSubmodule()
        _TESTREPOS = {'basic_annex':
                        {'network': 'https://github.com/datalad/testrepo--basic--r1',
                         'local': _basic_annex_test_repo.path,
                         'local-url': _basic_annex_test_repo.url},
                      'basic_git':
                        {'local': _basic_git_test_repo.path,
                         'local-url': _basic_git_test_repo.url},
                      'submodule_annex':
                        {'local': _submodule_annex_test_repo.path,
                         'local-url': _submodule_annex_test_repo.url},
                      'nested_submodule_annex':
                        {'local': _nested_submodule_annex_test_repo.path,
                         'local-url': _nested_submodule_annex_test_repo.url},
                      # TODO: append 'annex' to the name:
                      # Currently doesn't work with some annex tests, despite
                      # working manually. So, figure out how the tests' setup
                      # messes things up with this one.
                      'inner_submodule':
                        {'local': _inner_submodule_annex_test_repo.path,
                         'local-url': _inner_submodule_annex_test_repo.url}
                      }
        # assure that now we do have those test repos created -- delayed
        # their creation until actually used
        _basic_annex_test_repo.create()
        _basic_git_test_repo.create()
        _submodule_annex_test_repo.create()
        _nested_submodule_annex_test_repo.create()
        _inner_submodule_annex_test_repo.create()
    uris = []
    for name, spec in _TESTREPOS.items():
        if not re.match(regex, name):
            continue
        uris += [spec[x] for x in set(spec.keys()).intersection(flavors)]

        # additional flavors which might have not been
        if 'clone' in flavors and 'clone' not in spec:
            uris.append(clone_url(spec['local']))

        if 'network-clone' in flavors \
                and 'network' in spec \
                and 'network-clone' not in spec:
            uris.append(clone_url(spec['network']))

    return uris


def clone_url(url):
    runner = GitWitlessRunner()
    tdir = tempfile.mkdtemp(**get_tempfile_kwargs(
        {'dir': dl_cfg.get("datalad.tests.temp.dir")}, prefix='clone_url'))
    runner.run(["git", "clone", url, tdir], protocol=KillOutput)
    if GitRepo(tdir).is_with_annex():
        AnnexRepo(tdir, init=True)
    _TEMP_PATHS_CLONES.add(tdir)
    return tdir



def make_studyforrest_mockup(path):
    """Generate a dataset structure mimicking aspects of studyforrest.org

    Under the given path there are two directories:

    public - to be published datasets
    private - never to be published datasets

    The 'public' directory itself is a superdataset, the 'private' directory
    is just a directory that contains standalone datasets in subdirectories.
    """
    public = create(opj(path, 'public'), description="umbrella dataset")
    # the following tries to capture the evolution of the project
    phase1 = public.create('phase1',
                           description='old-style, no connection to RAW')
    structural = public.create('structural', description='anatomy')
    tnt = public.create('tnt', description='image templates')
    tnt.clone(source=phase1.path, path=opj('src', 'phase1'), reckless='auto')
    tnt.clone(source=structural.path, path=opj('src', 'structural'), reckless='auto')
    aligned = public.create('aligned', description='aligned image data')
    aligned.clone(source=phase1.path, path=opj('src', 'phase1'), reckless='auto')
    aligned.clone(source=tnt.path, path=opj('src', 'tnt'), reckless='auto')
    # new acquisition
    labet = create(opj(path, 'private', 'labet'), description="raw data ET")
    phase2_dicoms = create(opj(path, 'private', 'p2dicoms'), description="raw data P2MRI")
    phase2 = public.create('phase2',
                           description='new-style, RAW connection')
    phase2.clone(source=labet.path, path=opj('src', 'labet'), reckless='auto')
    phase2.clone(source=phase2_dicoms.path, path=opj('src', 'dicoms'), reckless='auto')
    # add to derivatives
    tnt.clone(source=phase2.path, path=opj('src', 'phase2'), reckless='auto')
    aligned.clone(source=phase2.path, path=opj('src', 'phase2'), reckless='auto')
    # never to be published media files
    media = create(opj(path, 'private', 'media'), description="raw data ET")
    # assuming all annotations are in one dataset (in reality this is also
    # a superdatasets with about 10 subdatasets
    annot = public.create('annotations', description='stimulus annotation')
    annot.clone(source=media.path, path=opj('src', 'media'), reckless='auto')
    # a few typical analysis datasets
    # (just doing 3, actual status quo is just shy of 10)
    # and also the real goal -> meta analysis
    metaanalysis = public.create('metaanalysis', description="analysis of analyses")
    for i in range(1, 3):
        ana = public.create('analysis{}'.format(i),
                            description='analysis{}'.format(i))
        ana.clone(source=annot.path, path=opj('src', 'annot'), reckless='auto')
        ana.clone(source=aligned.path, path=opj('src', 'aligned'), reckless='auto')
        ana.clone(source=tnt.path, path=opj('src', 'tnt'), reckless='auto')
        # link to metaanalysis
        metaanalysis.clone(source=ana.path, path=opj('src', 'ana{}'.format(i)),
                           reckless='auto')
        # simulate change in an input (but not raw) dataset
        create_tree(
            aligned.path,
            {'modification{}.txt'.format(i): 'unique{}'.format(i)})
        aligned.save()
    # finally aggregate data
    aggregate = public.create('aggregate', description='aggregate data')
    aggregate.clone(source=aligned.path, path=opj('src', 'aligned'), reckless='auto')
    # the toplevel dataset is intentionally left dirty, to reflect the
    # most likely condition for the joint dataset to be in at any given
    # point in time



#!/usr/bin/env python

import sys
from setuptools import setup
import versioneer

from _datalad_buildsupport.setup import (
    BuildManPage,
)

cmdclass = versioneer.get_cmdclass()
cmdclass.update(build_manpage=BuildManPage)

# Give setuptools a hint to complain if it's too old a version
# 43.0.0 allows us to put most metadata in setup.cfg and causes pyproject.toml
# to be automatically included in sdists
# Should match pyproject.toml
SETUP_REQUIRES = ['setuptools >= 43.0.0']
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

if __name__ == '__main__':
    setup(
        name='datalad_deprecated',
        version=versioneer.get_version(),
        cmdclass=cmdclass,
        setup_requires=SETUP_REQUIRES,
        entry_points={
            'datalad.extensions': [
                'deprecated=datalad_deprecated:command_suite',
            ],
            'datalad.metadata.extractors': [
                'annex=datalad_deprecated.metadata.extractors.annex:MetadataExtractor',
                'audio=datalad_deprecated.metadata.extractors.audio:MetadataExtractor',
                'datacite=datalad_deprecated.metadata.extractors.datacite:MetadataExtractor',
                'datalad_core=datalad_deprecated.metadata.extractors.datalad_core:MetadataExtractor',
                'datalad_rfc822=datalad_deprecated.metadata.extractors.datalad_rfc822:MetadataExtractor',
                'exif=datalad_deprecated.metadata.extractors.exif:MetadataExtractor',
                'frictionless_datapackage=datalad_deprecated.metadata.extractors.frictionless_datapackage:MetadataExtractor',
                'image=datalad_deprecated.metadata.extractors.image:MetadataExtractor',
                'xmp=datalad_deprecated.metadata.extractors.xmp:MetadataExtractor',
            ]
        },
    )



"""DataLad extension for deprecated functionality"""

__docformat__ = 'restructuredtext'


# Defines a datalad command suite.
# This variable must be bound as a setuptools entrypoint
# to be found by datalad
command_suite = (
    # description of the command suite, displayed in cmdline help
    "Deprecated functionality",
    [
        (
            'datalad_deprecated.ls',
            'Ls',
        ),
        (
            'datalad_deprecated.publish',
            'Publish',
        ),
        (
            'datalad_deprecated.annotate_paths',
            'AnnotatePaths',
        ),
        (
            'datalad_deprecated.metadata.search',
            'Search',
        ),
        (
            'datalad_deprecated.metadata.metadata',
            'Metadata',
        ),
        (
            'datalad_deprecated.metadata.extract_metadata',
            'ExtractMetadata',
            'extract-metadata',
            'extract_metadata',
        ),
        (
            'datalad_deprecated.metadata.aggregate',
            'AggregateMetaData',
            'aggregate-metadata',
            'aggregate_metadata',
        ),
    ]
)


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from datalad.support.constraints import EnsureChoice
from datalad.support.param import Parameter


reporton_opt = Parameter(
    args=('--reporton',),
    metavar='TYPE',
    doc="""choose on what type result to report on: 'datasets',
    'files', 'all' (both datasets and files), or 'none' (no report).""",
    constraints=EnsureChoice('all', 'datasets', 'files', 'none')
)
